const JSONRPC = require("jsonrpc-bidirectional");

module.exports = 
class AFCMakeCallEndpointBase extends JSONRPC.EndpointBase
{
	/**
	 * @param {string} strName
	 * @param {string} strPath
	 * @param {Object} objReflection
	 * @param {Class|null} classReverseCallsClient
	 */
	constructor(strName, strPath, objReflection, classReverseCallsClient)
	{
		super(strName, strPath, objReflection, classReverseCallsClient);

		this._objAFCPromises = {};
	}


	/**
	 * Proxy function which calls functions on this endpoint. 
	 * 
	 * It reuses promises for the same AFC ID to prevent duplicate calls 
	 * and to allow re-establishing a connection with BSIWebSocketServer's AFC service.
	 * 
	 * Failed promises are removed from the cache and are allowed to be retried.
	 * Successful promises remain in the cache for a while or forever, and are NOT allowed to be retried.
	 * 
	 * @param {JSONRPC.IncomingRequest} incomingRequest
	 * @param {number} nAFCID
	 * @param {string} strFunctionName
	 * @param {Array} arrParams
	 * 
	 * @returns {null}
	 */
	async afc_make_call(incomingRequest, nAFCID, strFunctionName, arrParams)
	{
		if(typeof nAFCID !== "number" || nAFCID <= 0)
		{
			throw new JSONRPC.Exception("nAFCID must be of type number and >= 1.", JSONRPC.Exception.INVALID_PARAMS);
		}

		if(typeof strFunctionName !== "string")
		{
			throw new JSONRPC.Exception("strFunctionName must be of type string.", JSONRPC.Exception.INVALID_PARAMS);
		}

		if(arrParams.constructor.name !== "Array")
		{
			throw new JSONRPC.Exception("arrParams must be of type Array.", JSONRPC.Exception.INVALID_PARAMS);
		}

		if(typeof this[strFunctionName] !== "function")
		{
			throw new JSONRPC.Exception("Method " + JSON.stringify(strFunctionName) + " not found on endpoint " + JSON.stringify(this.path) + ".", JSONRPC.Exception.METHOD_NOT_FOUND);
		}


		// @TODO: read mxResult (associated to nAFCID) from a persistent store. If found, return it.

		if(!this._objAFCPromises.hasOwnProperty(nAFCID))
		{
			this._objAFCPromises[nAFCID] = new Promise(async(fnResolve, fnReject) => 
			{
				try
				{
					const mxResult = await this[strFunctionName].apply(this, [incomingRequest].concat(arrParams));

					// @TODO: write mxResult to a persistent store, associated to nAFCID.

					// Waits only 5 minutes for BSI to ask this promise.
					setTimeout(
						() => 
						{
							if (this._objAFCPromises[nAFCID] != null)
							{
								delete this._objAFCPromises[nAFCID];
							}
						},
						5 * 60 * 1000 // 5 minutes
					);

					fnResolve(mxResult);
				}
				catch(error)
				{
					delete this._objAFCPromises[nAFCID];
					fnReject(error);
				}
			});
		}

		return this._objAFCPromises[nAFCID];
	}
};
